#include <config.h>

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <gnome.h>
#include <libgnomevfs/gnome-vfs.h>
#include <gdk-pixbuf/gdk-pixbuf.h>

/*
#define GIF_HACK
*/

#ifdef GIF_HACK
/* Private part of the GdkPixbuf structure */
struct _GdkPixbuf {
        GObject parent_instance;

	/* Color space */
	GdkColorspace colorspace;

	/* Number of channels, alpha included */
	int n_channels;

	/* Bits per channel */
	int bits_per_sample;

	/* Size */
	int width, height;

	/* Offset between rows */
	int rowstride;

	/* The pixel array */
	guchar *pixels;

	/* Destroy notification function; it is supposed to free the pixel array */
	GdkPixbufDestroyNotify destroy_fn;

	/* User data for the destroy notification function */
	gpointer destroy_fn_data;

	/* Do we have an alpha channel? */
	guint has_alpha : 1;
};
#endif

#define GIF_HACK

#ifdef GIF_HACK
/* Private part of the GdkPixbuf structure */
struct _GdkPixbuf {
        GObject parent_instance;

	/* Color space */
	GdkColorspace colorspace;

	/* Number of channels, alpha included */
	int n_channels;

	/* Bits per channel */
	int bits_per_sample;

	/* Size */
	int width, height;

	/* Offset between rows */
	int rowstride;

	/* The pixel array */
	guchar *pixels;

	/* Destroy notification function; it is supposed to free the pixel array */
	GdkPixbufDestroyNotify destroy_fn;

	/* User data for the destroy notification function */
	gpointer destroy_fn_data;

	/* Do we have an alpha channel? */
	guint has_alpha : 1;
};
#endif

#include "gtoonz.h"

static void
url_copy_cb (GtkWidget *widget,
	     gpointer data)
{
	Item *item = (Item *)data;
	GdkAtom atom;
	GtkClipboard *clipboard;

	atom = gdk_atom_intern ("CLIPBOARD", FALSE);
	clipboard = gtk_clipboard_get (atom);
	gtk_clipboard_set_text (clipboard, item->url, strlen(item->url));
}
/* file_*_cb ϵ忡 ϱ  Լ */
static void
file_close_cb (GnomeVFSAsyncHandle *handle,
	  GnomeVFSResult result,
	  gpointer data)
{
	return;
}
static void
file_read_cb (GnomeVFSAsyncHandle *handle,
	      GnomeVFSResult result,
	      gpointer buffer,
	      GnomeVFSFileSize bytes_requested,
	      GnomeVFSFileSize bytes_read,
	      gpointer data)
{
	LocalFile *file = (LocalFile *)data;
	GnomeVFSFileSize bytes;
	
	if (result != GNOME_VFS_OK) {
		if (result == GNOME_VFS_ERROR_EOF) {
			status_update (_("File download done"));
		} else {
			status_update (_("File download failed"));
		}
		gnome_vfs_close (file->fd);
		g_free (file->name);
		g_free (file->buf);
		g_free (file);
		gnome_vfs_async_close (handle, file_close_cb, NULL);
	} else {
		gdk_threads_enter ();
		gnome_vfs_write (file->fd, buffer, bytes_read, &bytes);
		gdk_threads_leave ();
		gnome_vfs_async_read (handle, buffer, BUF_SIZE,
				      (GnomeVFSAsyncReadCallback)file_read_cb, data);
	}
}
static void
file_open_cb (GnomeVFSAsyncHandle *handle,
	      GnomeVFSResult result,
	      gpointer data)
{
	LocalFile *file = (LocalFile *)data;
	
	gnome_vfs_async_read (handle, file->buf, BUF_SIZE,
			      (GnomeVFSAsyncReadCallback)file_read_cb, file);
}
static void
save_file_cb (GtkWidget *widget,
	      gpointer data)
{
	Item *item = (Item *)data;
	GnomeVFSAsyncHandle *handle;
	GnomeVFSResult result;
	LocalFile *file;

	file = g_new (LocalFile, 1);
	file->name = g_strdup_printf ("file:%s/%s",
				      g_get_home_dir (),
				      g_path_get_basename (item->url));
	result = gnome_vfs_create (&file->fd, file->name, GNOME_VFS_OPEN_WRITE, TRUE, 00600);
	if (result != GNOME_VFS_OK) {
		g_free (file->name);
		g_free (file);
		return;
	}
	file->buf = g_new0 (char, BUF_SIZE);

	status_update (_("File downloading... "));
	gnome_vfs_async_open(&handle, item->url,
			     GNOME_VFS_OPEN_READ, 0,
			     (GnomeVFSAsyncOpenCallback)file_open_cb, file);
}

static gboolean
button_press_cb (GtkWidget *widget,
		 GdkEventButton *event,
		 gpointer data)
{
	GtkWidget *menu;
	GtkWidget *button;

	if (event->type == GDK_BUTTON_PRESS && event->button == 3) {
		menu = gtk_menu_new ();

		button = gtk_menu_item_new_with_label (_("url copy"));
		g_signal_connect (G_OBJECT(button), "activate",
				  G_CALLBACK(url_copy_cb), data);
		gtk_menu_append (GTK_MENU(menu), button);
		gtk_widget_show (button);

		button = gtk_menu_item_new_with_label (_("save file"));
		g_signal_connect (G_OBJECT(button), "activate",
				  G_CALLBACK(save_file_cb), data);
		gtk_menu_append (GTK_MENU(menu), button);
		gtk_widget_show (button);

		gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL,
				event->button, event->time);
	}
	return TRUE;
}

static int
expose_cb(GtkWidget *da, GdkEvent *ev, gpointer data)
{
        GdkRectangle *r = &ev->expose.area;
        Item *item = data;
        assert(item->da == da);
        if (item->pixbuf) {
		int win_width, win_height;
		int pix_width;
		int pix_height;
		int xmargin, ymargin;
		int xmin, ymin;
		int xmax, ymax;
#ifdef GIF_HACK
		guint has_alpha;
#endif
		gdk_threads_enter ();
		pix_width = gdk_pixbuf_get_width(item->pixbuf);
		pix_height = gdk_pixbuf_get_height(item->pixbuf);
		gdk_window_get_size(da->window, &win_width, &win_height);
		xmin = r->x; ymin = r->y;
		xmax = r->x + r->width; ymax = r->y + r->height;
		xmargin = (win_width - pix_width) / 2;
		ymargin = (win_height - pix_height) / 2;

		if (xmin < xmargin) xmin = xmargin;
		if (ymin < ymargin) ymin = ymargin;
		if (xmax > xmargin + pix_width) xmax = xmargin + pix_width;
		if (ymax > ymargin + pix_height) ymax = ymargin + pix_height;
		if (xmax <= xmin || ymax <= ymin) {
			gdk_threads_leave();
			return 0;
		}

#ifdef GIF_HACK
		has_alpha = item->pixbuf->has_alpha;
		item->pixbuf->has_alpha = 0;
#endif

		gdk_pixbuf_render_to_drawable (item->pixbuf, da->window,
					       da->style->black_gc,
					       xmin - xmargin, ymin - ymargin,
					       xmin, ymin,
					       xmax - xmin, ymax - ymin,
					       GDK_RGB_DITHER_NORMAL, 0, 0);
#ifdef GIF_HACK
		item->pixbuf->has_alpha = has_alpha;
#endif
		gdk_threads_leave ();
	}
	return 1;
}

static void
progressive_prepared_cb (GdkPixbufLoader *loader,
			 gpointer data)
{
	GdkPixbuf *pixbuf;
	Item *item = data;
	int width, height;

	pixbuf = gdk_pixbuf_loader_get_pixbuf (loader);
	gdk_pixbuf_fill (pixbuf, 0xaaaaaaff);
	width = gdk_pixbuf_get_width (pixbuf);
	height = gdk_pixbuf_get_height (pixbuf);
	gtk_drawing_area_size (GTK_DRAWING_AREA(item->da), width, height);
	item->pixbuf = pixbuf;
}

static void
progressive_update_cb (GdkPixbufLoader *loader,
		       gint x,
		       gint y,
		       gint width,
		       gint height,
		       gpointer data)
{
	Item *item = data;
	int win_width, win_height;
	int xmargin, ymargin;
	int pix_width = gdk_pixbuf_get_width(item->pixbuf);
	int pix_height = gdk_pixbuf_get_height(item->pixbuf);
#ifdef GIF_HACK
	guint has_alpha;
#endif
	if (!height || !GTK_WIDGET_REALIZED(item->da)) return;
	gdk_window_get_size(item->da->window, &win_width, &win_height);
	xmargin = (win_width - pix_width) / 2;
	ymargin = (win_height - pix_height) / 2;
	if (ymargin < 0 || xmargin < 0) {
	        return;
	}
#ifdef GIF_HACK
	has_alpha = item->pixbuf->has_alpha;
	item->pixbuf->has_alpha = 0;
#endif
	gdk_pixbuf_render_to_drawable (item->pixbuf, item->da->window,
				       item->da->style->black_gc,
				       x, y, x + xmargin, y + ymargin,
				       width, height,
				       GDK_RGB_DITHER_NORMAL, 0, 0);
#ifdef GIF_HACK
	item->pixbuf->has_alpha = has_alpha;
#endif
}

static void
close_cb (GnomeVFSAsyncHandle *handle,
	  GnomeVFSResult result,
	  gpointer data)
{
	Item *item = (Item *)data;
	gdk_threads_enter ();
	
	gdk_window_set_cursor (item->cartoon->app->window, 0);
	
	gdk_threads_leave ();
	return;
}

static void
read_cb (GnomeVFSAsyncHandle *handle,
	 GnomeVFSResult result,
	 gpointer buffer,
	 GnomeVFSFileSize bytes_requested,
	 GnomeVFSFileSize bytes_read,
	 gpointer data)
{
	Item *item = (Item *)data;
	assert (item->handle == handle);
	gdk_threads_enter ();

	if (result != GNOME_VFS_OK) {
		if (result == GNOME_VFS_ERROR_EOF) {
			status_update (_("Done"));
		} else {
			status_update (_("Download failed"));
		}
		item->handle = 0;
		gnome_vfs_async_close (handle, close_cb, item);
	} else {
		gdk_pixbuf_loader_write (item->loader, buffer, bytes_read, NULL);
		gnome_vfs_async_read (handle, buffer, BUF_SIZE,
				      (GnomeVFSAsyncReadCallback)read_cb, item);
	}
	gdk_threads_leave ();
	return;
}

static void
open_cb (GnomeVFSAsyncHandle *handle,
	 GnomeVFSResult result,
	 gpointer data)
{
	Item *item = (Item *)data;
 	gdk_threads_enter ();
	
	if (result != GNOME_VFS_OK) {
		status_update (_("file not found"));
		item->handle = 0;
		gdk_window_set_cursor (item->cartoon->app->window, 0);
	} else {
		gnome_vfs_async_read (handle, item->buf, BUF_SIZE,
				      (GnomeVFSAsyncReadCallback)read_cb, item);
	}
 	gdk_threads_leave ();
	return;
}

void
download (Cartoon *cartoon)
{
	GSList *list;
	Item *item;
	GtkWidget *image;
	GdkCursor *cursor;


	list = cartoon->item_list;
	
	status_update (_("Downloading..."));
	cursor = gdk_cursor_new (GDK_WATCH);
	gdk_window_set_cursor (cartoon->app->window, cursor);
	gdk_cursor_unref (cursor);
	
	while (list) {
		item = list->data;

		item->cartoon = cartoon;
 		item->da = gtk_drawing_area_new ();
 		gtk_widget_add_events (item->da, GDK_BUTTON_PRESS_MASK);
 		gtk_box_pack_start (GTK_BOX(cartoon->images_box), item->da,
 				    TRUE, TRUE, 0);
 		g_signal_connect (G_OBJECT(item->da), "button-press-event",
 				  G_CALLBACK(button_press_cb), item);
 		g_signal_connect (G_OBJECT(item->da), "expose-event",
 				  G_CALLBACK(expose_cb), item);
 		gtk_widget_show (item->da);

		item->buf = g_new(char, BUF_SIZE);
		item->loader = gdk_pixbuf_loader_new ();
		g_signal_connect (G_OBJECT(item->loader), "area_prepared",
				  G_CALLBACK(progressive_prepared_cb), item);
		g_signal_connect (G_OBJECT(item->loader), "area_updated",
				  G_CALLBACK(progressive_update_cb), item);

		gnome_vfs_async_open(&item->handle, item->url,
				     GNOME_VFS_OPEN_READ, -5,
				     (GnomeVFSAsyncOpenCallback)open_cb, item);
		list = list->next;
	}
	return;
}
